import os
import tempfile
import zlib
from contextlib import contextmanager

from dagster import Field, StringSource, check, seven
from dagster.core.storage.compute_log_manager import (
    MAX_BYTES_FILE_READ,
    ComputeIOType,
    ComputeLogManager,
)
from dagster.core.storage.local_compute_log_manager import LocalComputeLogManager
from dagster.serdes import ConfigurableClass, ConfigurableClassData
from dagster.utils import ensure_file

from ...errors import raise_http_error


class CloudComputeLogManager(ComputeLogManager, ConfigurableClass):
    def __init__(
        self,
        local_dir=None,
        inst_data=None,
    ):

        # proxy calls to local compute log manager (for subscriptions, etc)
        if not local_dir:
            local_dir = seven.get_system_temp_directory()

        self.local_manager = LocalComputeLogManager(local_dir)
        self._inst_data = check.opt_inst_param(inst_data, "inst_data", ConfigurableClassData)

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {
            "local_dir": Field(StringSource, is_required=False),
        }

    @staticmethod
    def from_config_value(inst_data, config_value):
        return CloudComputeLogManager(inst_data=inst_data, **config_value)

    @contextmanager
    def _watch_logs(self, pipeline_run, step_key=None):
        # proxy watching to the local compute log manager, interacting with the filesystem
        with self.local_manager._watch_logs(  # pylint: disable=protected-access
            pipeline_run, step_key
        ):
            yield

    def get_local_path(self, run_id, key, io_type):
        return self.local_manager.get_local_path(run_id, key, io_type)

    def on_watch_start(self, pipeline_run, step_key):
        self.local_manager.on_watch_start(pipeline_run, step_key)

    def is_watch_completed(self, run_id, key):
        return self.local_manager.is_watch_completed(run_id, key)

    def on_watch_finish(self, pipeline_run, step_key):
        self.local_manager.on_watch_finish(pipeline_run, step_key)
        key = self.local_manager.get_key(pipeline_run, step_key)
        self._upload_from_local(pipeline_run.run_id, key, ComputeIOType.STDOUT)
        self._upload_from_local(pipeline_run.run_id, key, ComputeIOType.STDERR)

    def _upload_from_local(self, run_id, key, io_type):
        path = self.get_local_path(run_id, key, io_type)
        ensure_file(path)
        with open(path, "rb") as data:
            with tempfile.TemporaryDirectory() as temp_dir:
                dst = os.path.join(temp_dir, "compute_log.tmp")
                with open(dst, "wb") as compressed:
                    compressed.write(zlib.compress(data.read()))

                with open(dst, "rb") as compressed:
                    resp = self._instance.requests_session.post(
                        self._instance.dagster_cloud_upload_logs_url,
                        headers=self._instance.dagster_cloud_api_headers,
                        params={
                            "run_id": run_id,
                            "key": key,
                            "io_type": io_type.value,
                            "compressed": True,
                        },
                        files={"compute_log.tmp": compressed},
                        timeout=self._instance.dagster_cloud_api_timeout,
                    )
        raise_http_error(resp)

    def download_url(self, run_id, key, io_type):
        raise NotImplementedError("User Agent should not need to download compute logs")

    def read_logs_file(self, run_id, key, io_type, cursor=0, max_bytes=MAX_BYTES_FILE_READ):
        raise NotImplementedError("User Agent should not need to download compute logs")

    def on_subscribe(self, subscription):
        raise NotImplementedError("User Agent should not need to download compute logs")



import logging
from contextlib import contextmanager
from typing import Dict, Iterable, Optional

import kubernetes
import kubernetes.client as client
from dagster import Array, BoolSource, Field, Noneable, Permissive, StringSource, check
from dagster.config.field_utils import Shape
from dagster.core.executor.step_delegating import StepHandler
from dagster.core.host_representation.grpc_server_registry import GrpcServerEndpoint
from dagster.serdes import ConfigurableClass
from dagster.utils import ensure_single_item
from dagster_cloud.execution.watchful_run_launcher.k8s import WatchfulK8sRunLauncher
from dagster_cloud.executor import (
    DAGSTER_CLOUD_EXECUTOR_K8S_CONFIG_KEY,
    DAGSTER_CLOUD_EXECUTOR_NAME,
)
from dagster_cloud.workspace.origin import CodeDeploymentMetadata
from dagster_k8s.executor import K8sStepHandler
from dagster_k8s.job import DagsterK8sJobConfig
from kubernetes.client.rest import ApiException

from ..user_code_launcher import ReconcileUserCodeLauncher
from .utils import (
    SERVICE_PORT,
    construct_repo_location_deployment,
    construct_repo_location_service,
    get_unique_label_for_location,
    unique_resource_name,
    wait_for_deployment_complete,
)

DEPLOYMENT_TIMEOUT = 90  # Can take time to pull images
SERVER_TIMEOUT = 60


class K8sUserCodeLauncher(ReconcileUserCodeLauncher[str], ConfigurableClass):
    def __init__(
        self,
        dagster_home,
        instance_config_map,
        inst_data=None,
        namespace=None,
        kubeconfig_file=None,
        pull_policy=None,
        env_config_maps=None,
        env_secrets=None,
        service_account_name=None,
        volume_mounts=None,
        volumes=None,
        image_pull_secrets=None,
    ):
        self._inst_data = inst_data
        self._logger = logging.getLogger("K8sUserCodeLauncher")
        self._dagster_home = check.str_param(dagster_home, "dagster_home")
        self._instance_config_map = check.str_param(instance_config_map, "instance_config_map")
        self._namespace = namespace
        self._pull_policy = pull_policy
        self._env_config_maps = check.opt_list_param(
            env_config_maps, "env_config_maps", of_type=str
        )
        self._env_secrets = check.opt_list_param(env_secrets, "env_secrets", of_type=str)
        self._service_account_name = check.opt_str_param(
            service_account_name, "service_account_name"
        )
        self._volume_mounts = check.opt_list_param(volume_mounts, "volume_mounts")
        self._volumes = check.opt_list_param(volumes, "volumes")
        self._image_pull_secrets = check.opt_list_param(
            image_pull_secrets, "image_pull_secrets", of_type=dict
        )

        if kubeconfig_file:
            kubernetes.config.load_kube_config(kubeconfig_file)
        else:
            kubernetes.config.load_incluster_config()

        super(K8sUserCodeLauncher, self).__init__()

        self._launcher = WatchfulK8sRunLauncher(
            dagster_home=self._dagster_home,
            instance_config_map=self._instance_config_map,
            postgres_password_secret=None,
            job_image=None,
            image_pull_policy=self._pull_policy,
            image_pull_secrets=self._image_pull_secrets,
            service_account_name=self._service_account_name,
            env_config_maps=self._env_config_maps,
            env_secrets=self._env_secrets,
            job_namespace=self._namespace,
            volume_mounts=self._volume_mounts,
            volumes=self._volumes,
        )

    @property
    def requires_images(self):
        return True

    def register_instance(self, instance):
        super().register_instance(instance)
        self._launcher.register_instance(instance)

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {
            "dagster_home": Field(StringSource, is_required=True),
            "instance_config_map": Field(StringSource, is_required=True),
            "namespace": Field(StringSource, is_required=False, default_value="default"),
            "kubeconfig_file": Field(StringSource, is_required=False),
            "pull_policy": Field(StringSource, is_required=False, default_value="Always"),
            "env_config_maps": Field(
                Noneable(Array(StringSource)),
                is_required=False,
                description="A list of custom ConfigMapEnvSource names from which to draw "
                "environment variables (using ``envFrom``) for the Job. Default: ``[]``. See:"
                "https://kubernetes.io/docs/tasks/inject-data-application/define-environment-variable-container/#define-an-environment-variable-for-a-container",
            ),
            "env_secrets": Field(
                Noneable(Array(StringSource)),
                is_required=False,
                description="A list of custom Secret names from which to draw environment "
                "variables (using ``envFrom``) for the Job. Default: ``[]``. See:"
                "https://kubernetes.io/docs/tasks/inject-data-application/distribute-credentials-secure/#configure-all-key-value-pairs-in-a-secret-as-container-environment-variables",
            ),
            "service_account_name": Field(
                Noneable(StringSource),
                is_required=False,
                description="Override the name of the Kubernetes service account under "
                "which to run.",
            ),
            "volume_mounts": Field(
                Array(
                    Shape(
                        {
                            "name": StringSource,
                            "mountPath": StringSource,
                            "mountPropagation": Field(StringSource, is_required=False),
                            "readOnly": Field(BoolSource, is_required=False),
                            "subPath": Field(StringSource, is_required=False),
                            "subPathExpr": Field(StringSource, is_required=False),
                        }
                    )
                ),
                is_required=False,
                default_value=[],
                description="A list of volume mounts to include in the job's container. Default: ``[]``. See: "
                "https://v1-18.docs.kubernetes.io/docs/reference/generated/kubernetes-api/v1.18/#volumemount-v1-core",
            ),
            "volumes": Field(
                Array(
                    Permissive(
                        {
                            "name": str,
                        }
                    )
                ),
                is_required=False,
                default_value=[],
                description="A list of volumes to include in the Job's Pod. Default: ``[]``. For the many "
                "possible volume source types that can be included, see: "
                "https://v1-18.docs.kubernetes.io/docs/reference/generated/kubernetes-api/v1.18/#volume-v1-core",
            ),
            "image_pull_secrets": Field(
                Noneable(Array(Shape({"name": StringSource}))),
                is_required=False,
                description="Specifies that Kubernetes should get the credentials from "
                "the Secrets named in this list.",
            ),
        }

    @staticmethod
    def from_config_value(inst_data, config_value):
        return K8sUserCodeLauncher(
            dagster_home=config_value.get("dagster_home"),
            instance_config_map=config_value.get("instance_config_map"),
            inst_data=inst_data,
            namespace=config_value.get("namespace"),
            kubeconfig_file=config_value.get("kubeconfig_file"),
            pull_policy=config_value.get("pull_policy"),
            env_config_maps=config_value.get("env_config_maps", []),
            env_secrets=config_value.get("env_secrets", []),
            service_account_name=config_value.get("service_account_name"),
            image_pull_secrets=config_value.get("image_pull_secrets"),
            volume_mounts=config_value.get("volume_mounts"),
            volumes=config_value.get("volumes"),
        )

    @contextmanager
    def _get_api_instance(self):
        with client.ApiClient() as api_client:  # pylint: disable=not-context-manager
            yield client.AppsV1Api(api_client)

    def _create_new_server_endpoint(
        self, location_name: str, metadata: CodeDeploymentMetadata
    ) -> GrpcServerEndpoint:
        resource_name = unique_resource_name(location_name)

        try:
            with self._get_api_instance() as api_instance:
                api_response = api_instance.create_namespaced_deployment(
                    self._namespace,
                    construct_repo_location_deployment(
                        location_name,
                        resource_name,
                        metadata,
                        self._pull_policy,
                        self._env_config_maps,
                        self._env_secrets,
                        self._service_account_name,
                        self._image_pull_secrets,
                        self._volume_mounts,
                        self._volumes,
                    ),
                )
            self._logger.info("Created deployment: {}".format(api_response.metadata.name))
        except ApiException as e:
            self._logger.error(
                "Exception when calling AppsV1Api->create_namespaced_deployment: %s\n" % e
            )
            raise e

        try:
            api_response = client.CoreV1Api().create_namespaced_service(
                self._namespace,
                construct_repo_location_service(location_name, resource_name),
            )
            self._logger.info("Created service: {}".format(api_response.metadata.name))
        except ApiException as e:
            self._logger.error(
                "Exception when calling AppsV1Api->create_namespaced_service: %s\n" % e
            )
            raise e

        wait_for_deployment_complete(
            resource_name,
            self._namespace,
            self._logger,
            location_name,
            metadata,
            existing_pods=[],
            timeout=DEPLOYMENT_TIMEOUT,
        )
        server_id = self._wait_for_server(
            host=resource_name, port=SERVICE_PORT, timeout=SERVER_TIMEOUT
        )

        endpoint = GrpcServerEndpoint(
            server_id=server_id,
            host=resource_name,
            port=SERVICE_PORT,
            socket=None,
        )

        return endpoint

    def _get_server_handles_for_location(self, location_name: str) -> Iterable[str]:
        with self._get_api_instance() as api_instance:
            deployments = api_instance.list_namespaced_deployment(
                self._namespace,
                label_selector=f"location_hash={get_unique_label_for_location(location_name)}",
            ).items
            return [deployment.metadata.name for deployment in deployments]

    def _cleanup_servers(self) -> None:
        with self._get_api_instance() as api_instance:
            deployments = api_instance.list_namespaced_deployment(
                self._namespace,
                label_selector="managed_by=K8sUserCodeLauncher",
            ).items
            for deployment in deployments:
                self._remove_server_handle(deployment.metadata.name)

    def _remove_server_handle(self, server_handle: str) -> None:
        check.str_param(server_handle, "server_handle")
        with self._get_api_instance() as api_instance:
            api_instance.delete_namespaced_deployment(server_handle, self._namespace)
        client.CoreV1Api().delete_namespaced_service(server_handle, self._namespace)
        self._logger.info("Removed deployment and service: {}".format(server_handle))

    def __exit__(self, exception_type, exception_value, traceback):
        super().__exit__(exception_value, exception_value, traceback)
        self._launcher.dispose()

    def get_step_handler(self, execution_config: Optional[Dict]) -> StepHandler:
        executor_name, exc_cfg = (
            ensure_single_item(execution_config) if execution_config else (None, {})
        )

        k8s_cfg = {}

        if executor_name == DAGSTER_CLOUD_EXECUTOR_NAME:
            k8s_cfg = exc_cfg.get(DAGSTER_CLOUD_EXECUTOR_K8S_CONFIG_KEY, {})

        return K8sStepHandler(
            job_config=DagsterK8sJobConfig(
                dagster_home=self._dagster_home,
                instance_config_map=self._instance_config_map,
                postgres_password_secret=None,
                job_image=None,
                image_pull_policy=(
                    k8s_cfg.get("image_pull_policy")
                    if k8s_cfg.get("image_pull_policy") != None
                    else self._pull_policy
                ),
                image_pull_secrets=self._image_pull_secrets
                + (k8s_cfg.get("image_pull_secrets") or []),
                service_account_name=(
                    k8s_cfg.get("service_account_name")
                    if k8s_cfg.get("service_account_name") != None
                    else self._service_account_name
                ),
                env_config_maps=self._env_config_maps + (k8s_cfg.get("env_config_maps") or []),
                env_secrets=self._env_secrets + (k8s_cfg.get("env_secrets") or []),
                volume_mounts=self._volume_mounts + (k8s_cfg.get("volume_mounts") or []),
                volumes=self._volumes + (k8s_cfg.get("volumes") or []),
            ),
            job_namespace=(
                k8s_cfg.get("job_namespace")
                if k8s_cfg.get("job_namespace") != None
                else self._namespace
            ),
            load_incluster_config=True,
            kubeconfig_file=None,
        )

    def run_launcher(self):
        return self._launcher



import time
from pathlib import Path
from typing import Any, Dict
from dagster_cloud.storage.client import GqlShimClient

import yaml
from dagster import check
from dagster.serdes.serdes import deserialize_json_to_dagster_namedtuple
from dagster.utils import DEFAULT_WORKSPACE_YAML_FILENAME
from dagster_cloud.workspace.config_schema import process_workspace_config
from dagster_cloud.workspace.origin import CodeDeploymentMetadata
from typer import Argument, Option, Typer

from ...cli import gql, ui
from ...cli.config_utils import dagster_cloud_options
from ...cli.utils import add_options

app = Typer(help="Commands for managing a dagster cloud workspace.")

_DEPLOYMENT_METADATA_OPTIONS = {
    "image": (str, Option(None, "--image", help="Docker image.")),
    "python_file": (
        Path,
        Option(
            None, "--python-file", "-f", exists=False, help="Python file where repository lives."
        ),
    ),
    "package_name": (
        str,
        Option(None, "--package-name", "-p", help="Python package where repositories live"),
    ),
}


def _get_location_input(location: str, kwargs: Dict[str, Any]) -> gql.CliInputCodeLocation:
    python_file = kwargs.get("python_file")
    return gql.CliInputCodeLocation(
        name=location,
        python_file=str(python_file) if python_file else None,
        package_name=kwargs.get("package_name"),
        image=kwargs.get("image"),
    )


def _add_or_update_location(client: GqlShimClient, location_data: gql.CliInputCodeLocation) -> None:
    try:
        gql.add_or_update_code_location(client, location_data)
        ui.print(f"Added or updated location {location_data.name}.")
        wait_for_load(client, [location_data.name])
    except Exception as e:
        raise ui.error(str(e))


@app.command(name="add-location")
@dagster_cloud_options(allow_empty=True, requires_url=True)
@add_options(_DEPLOYMENT_METADATA_OPTIONS)
def add_command(
    api_token: str,
    url: str,
    location: str = Argument(..., help="Code location name."),
    **kwargs,
):
    """Add or update the image for a repository location in the workspace."""
    client = gql.graphql_client_from_url(url, api_token)
    location_data = _get_location_input(location, kwargs)
    _add_or_update_location(client, location_data)


@app.command(
    name="update-location",
)
@dagster_cloud_options(allow_empty=True, requires_url=True)
@add_options(_DEPLOYMENT_METADATA_OPTIONS)
def update_command(
    api_token: str,
    url: str,
    location: str = Argument(..., help="Code location name."),
    **kwargs,
):
    """Update the image for a repository location in the workspace."""
    client = gql.graphql_client_from_url(url, api_token)
    location_data = _get_location_input(location, kwargs)
    _add_or_update_location(client, location_data)


def wait_for_load(client, locations):
    start_time = time.time()
    ui.print(f"Waiting for agent to sync changes to {','.join(locations)}...")
    while True:
        if time.time() - start_time > 300:
            raise ui.error("Timed out waiting for location data to update")

        try:
            nodes = gql.fetch_code_locations(client)
        except Exception as e:
            raise ui.error(str(e))

        nodes_by_location = {node["name"]: node for node in nodes}

        if all(
            location in nodes_by_location
            and nodes_by_location[location].get("loadStatus") == "LOADED"
            for location in locations
        ):

            error_locations = [
                location
                for location in locations
                if "locationOrLoadError" in nodes_by_location[location]
                and nodes_by_location[location]["locationOrLoadError"]["__typename"]
                == "PythonError"
            ]

            if error_locations:
                error_string = "Some locations failed to load after being synced by the agent:\n" + "\n".join(
                    [
                        f"Error loading {error_location}: {str(nodes_by_location[error_location]['locationOrLoadError'])}"
                        for error_location in error_locations
                    ]
                )
                raise ui.error(error_string)
            else:
                ui.print(
                    f"Agent synced changes to {','.join(locations)}. Changes should now be visible in dagit."
                )
                break

        time.sleep(3)


@app.command(
    name="delete-location",
)
@dagster_cloud_options(allow_empty=True, requires_url=True)
def delete_command(
    api_token: str,
    url: str,
    location: str = Argument(..., help="Code location name."),
):
    """Delete a repository location from the workspace."""
    client = gql.graphql_client_from_url(url, api_token)
    try:
        gql.delete_code_location(client, location)
        ui.print(f"Deleted location {location}.")
    except Exception as e:
        raise ui.error(str(e))


@app.command(
    name="list",
)
@dagster_cloud_options(allow_empty=True, requires_url=True)
def list_command(
    url: str,
    api_token: str,
):
    """List repository locations in the workspace."""
    client = gql.graphql_client_from_url(url, api_token)
    execute_list_command(client)


def execute_list_command(client):
    list_res = gql.fetch_workspace_entries(client)

    ui.print("Listing locations...")

    for location in list_res:
        metadata = check.inst(
            deserialize_json_to_dagster_namedtuple(location["serializedDeploymentMetadata"]),
            CodeDeploymentMetadata,
        )

        location_desc = [location["locationName"]]
        if metadata.python_file:
            location_desc.append(f"File: {metadata.python_file}")
        if metadata.package_name:
            location_desc.append(f"Package: {metadata.package_name}")
        if metadata.image:
            location_desc.append(f"Image: {metadata.image}")

        ui.print("\t".join(location_desc))


@app.command(name="sync")
@dagster_cloud_options(allow_empty=True, requires_url=True)
def sync_command(
    url: str,
    api_token: str,
    workspace: Path = Option(
        DEFAULT_WORKSPACE_YAML_FILENAME,
        "--workspace",
        "-w",
        exists=True,
        help="Path to workspace file.",
    ),
):
    """"Sync the workspace with the contents of a workspace.yaml file."""
    client = gql.graphql_client_from_url(url, api_token)
    execute_sync_command(client, workspace)


def execute_sync_command(client, workspace):
    with open(str(workspace), "r") as f:
        config = yaml.load(f.read(), Loader=yaml.SafeLoader)
        process_workspace_config(config)

    try:
        locations = gql.reconcile_code_locations(
            client,
            [
                gql.CliInputCodeLocation(
                    name=name,
                    image=m.get("image"),
                    package_name=m.get("package_name"),
                    python_file=m.get("python_file"),
                )
                for name, m in config["locations"].items()
            ],
        )
        ui.print(f"Synced locations: {', '.join(locations)}")
        wait_for_load(client, locations)
    except Exception as e:
        raise ui.error(str(e))



import os
import sys
import tempfile
import threading
import time
import zlib
from abc import abstractmethod, abstractproperty
from typing import (
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Set,
    TypeVar,
    Union,
)

from dagster import check
from dagster.api.get_server_id import sync_get_server_id
from dagster.api.list_repositories import sync_list_repositories_grpc
from dagster.core.executor.step_delegating import StepHandler
from dagster.core.host_representation import ExternalRepositoryOrigin
from dagster.core.host_representation.grpc_server_registry import (
    GrpcServerEndpoint,
    GrpcServerRegistry,
)
from dagster.core.host_representation.origin import (
    RegisteredRepositoryLocationOrigin,
    RepositoryLocationOrigin,
)
from dagster.core.instance import MayHaveInstanceWeakref
from dagster.daemon.daemon import get_default_daemon_logger
from dagster.grpc.client import DagsterGrpcClient
from dagster.grpc.types import GetCurrentImageResult
from dagster.serdes import deserialize_as, serialize_dagster_namedtuple, whitelist_for_serdes
from dagster.utils.error import SerializableErrorInfo, serializable_error_info_from_exc_info
from dagster_cloud.api.dagster_cloud_api import (
    DagsterCloudUploadLocationData,
    DagsterCloudUploadRepositoryData,
    DagsterCloudUploadWorkspaceEntry,
)
from dagster_cloud.errors import raise_http_error
from dagster_cloud.execution.watchful_run_launcher.base import WatchfulRunLauncher
from dagster_cloud.util import diff_serializable_namedtuple_map
from dagster_cloud.workspace.origin import CodeDeploymentMetadata

USER_CODE_LAUNCHER_RECONCILE_INTERVAL = 1

ServerHandle = TypeVar("ServerHandle")


@whitelist_for_serdes
class UserCodeLauncherEntry(
    NamedTuple(
        "_UserCodeLauncherEntry",
        [("code_deployment_metadata", CodeDeploymentMetadata), ("update_timestamp", float)],
    )
):
    pass


class DagsterCloudUserCodeLauncher(GrpcServerRegistry, MayHaveInstanceWeakref):
    def __init__(self):
        self._logger = get_default_daemon_logger("DagsterUserCodeLauncher")
        self._started: bool = False

    def start(self):
        check.invariant(
            not self._started,
            "Called start() on a DagsterCloudUserCodeLauncher that was already started",
        )
        # Begin spinning user code up and down
        self._started = True

    @abstractmethod
    def update_grpc_metadata(
        self,
        desired_metadata: Dict[str, UserCodeLauncherEntry],
        upload_locations: Set[str],
    ):
        pass

    def supports_origin(self, repository_location_origin: RepositoryLocationOrigin) -> bool:
        return isinstance(repository_location_origin, RegisteredRepositoryLocationOrigin)

    @abstractproperty
    def is_workspace_ready(self) -> bool:
        pass

    @property
    def supports_reload(self) -> bool:
        return False

    def reload_grpc_endpoint(self, repository_location_origin: RepositoryLocationOrigin):
        raise NotImplementedError("Call update_grpc_metadata to update gRPC endpoints")

    @abstractmethod
    def _get_repository_location_origin(self, location_name: str) -> RepositoryLocationOrigin:
        pass

    def _update_workspace_entry(self, workspace_entry: DagsterCloudUploadWorkspaceEntry):
        with tempfile.TemporaryDirectory() as temp_dir:
            dst = os.path.join(temp_dir, "workspace_entry.tmp")
            with open(dst, "wb") as f:
                f.write(
                    zlib.compress(serialize_dagster_namedtuple(workspace_entry).encode("utf-8"))
                )

            with open(dst, "rb") as f:
                self._logger.info(
                    "Uploading workspace entry for {location_name} ({size} bytes)".format(
                        location_name=workspace_entry.location_name, size=os.path.getsize(dst)
                    )
                )

                resp = self._instance.requests_session.post(
                    self._instance.dagster_cloud_upload_workspace_entry_url,
                    headers=self._instance.dagster_cloud_api_headers,
                    files={"workspace_entry.tmp": f},
                    timeout=self._instance.dagster_cloud_api_timeout,
                )
                raise_http_error(resp)

                self._logger.info(
                    "Successfully uploaded workspace entry for {location_name}".format(
                        location_name=workspace_entry.location_name
                    )
                )

    def _get_upload_location_data(self, location_name: str) -> DagsterCloudUploadLocationData:
        location_origin = self._get_repository_location_origin(location_name)
        client = self.get_grpc_endpoint(location_origin).create_client()

        list_repositories_response = sync_list_repositories_grpc(client)

        upload_repo_datas: List[DagsterCloudUploadRepositoryData] = []

        for (
            repository_name,
            code_pointer,
        ) in list_repositories_response.repository_code_pointer_dict.items():
            external_repository_chunks = list(
                client.streaming_external_repository(
                    external_repository_origin=ExternalRepositoryOrigin(
                        location_origin,
                        repository_name,
                    )
                )
            )

            serialized_repository_data = "".join(
                [
                    chunk["serialized_external_repository_chunk"]
                    for chunk in external_repository_chunks
                ]
            )
            # Don't deserialize in case there are breaking changes - let the server do it
            upload_repo_datas.append(
                DagsterCloudUploadRepositoryData(
                    repository_name=repository_name,
                    code_pointer=code_pointer,
                    serialized_repository_data=serialized_repository_data,
                )
            )

        return DagsterCloudUploadLocationData(
            upload_repository_datas=upload_repo_datas,
            container_image=deserialize_as(
                client.get_current_image(), GetCurrentImageResult
            ).current_image,
            executable_path=list_repositories_response.executable_path,
        )

    def _update_location_error(
        self,
        location_name: str,
        error_info: SerializableErrorInfo,
        metadata: CodeDeploymentMetadata,
    ):
        self._logger.error(
            "Unable to load location {location_name}. Updating location with error data: {error_info}.".format(
                location_name=location_name,
                error_info=str(error_info),
            )
        )

        # Update serialized error
        errored_workspace_entry = DagsterCloudUploadWorkspaceEntry(
            location_name=location_name,
            deployment_metadata=metadata,
            upload_location_data=None,
            serialized_error_info=error_info,
        )

        self._update_workspace_entry(errored_workspace_entry)

    def _update_location_data(self, location_name: str, endpoint, metadata: CodeDeploymentMetadata):
        self._logger.info(
            "Uploading metadata for location {location_name}".format(location_name=location_name)
        )

        try:
            if isinstance(endpoint, SerializableErrorInfo):
                self._update_location_error(location_name, error_info=endpoint, metadata=metadata)
                return

            try:
                loaded_workspace_entry = DagsterCloudUploadWorkspaceEntry(
                    location_name=location_name,
                    deployment_metadata=metadata,
                    upload_location_data=self._get_upload_location_data(location_name),
                    serialized_error_info=None,
                )
            except Exception:
                self._update_location_error(
                    location_name,
                    error_info=serializable_error_info_from_exc_info(sys.exc_info()),
                    metadata=metadata,
                )
                return

            self._logger.info(
                "Updating location {location_name} with repository load data".format(
                    location_name=location_name,
                )
            )
            self._update_workspace_entry(loaded_workspace_entry)
        except Exception:
            # Don't let a failure uploading the serialized data keep other locations
            # from being updated or continue reconciling in a loop
            error_info = serializable_error_info_from_exc_info(sys.exc_info())
            self._logger.error(
                "Error while writing location data for location {location_name}: {error_info}".format(
                    location_name=location_name,
                    error_info=error_info,
                )
            )


class ReconcileUserCodeLauncher(DagsterCloudUserCodeLauncher, Generic[ServerHandle]):
    def __init__(self):
        self._grpc_endpoints: Dict[str, Union[GrpcServerEndpoint, SerializableErrorInfo]] = {}
        self._grpc_endpoints_lock = threading.Lock()
        self._logger = get_default_daemon_logger("ReconcileUserCodeLauncher")

        # periodically reconciles to make desired = actual
        self._desired_entries: Optional[Dict[str, UserCodeLauncherEntry]] = None
        self._actual_entries = {}
        self._metadata_lock = threading.Lock()

        self._upload_locations: Set[str] = set()

        super().__init__()

        self._reconcile_count = 0
        self._reconcile_grpc_metadata_shutdown_event = threading.Event()
        self._reconcile_grpc_metadata_thread = None

        self._is_workspace_ready_lock = threading.Lock()
        self._is_workspace_ready: bool = False

    @property
    def is_workspace_ready(self) -> bool:
        with self._is_workspace_ready_lock:
            return self._is_workspace_ready

    @abstractproperty
    def requires_images(self) -> bool:
        pass

    @abstractmethod
    def _get_server_handles_for_location(self, location_name: str) -> Iterable[ServerHandle]:
        """Return a list of 'handles' that represent all running servers for a given location.
        Typically this will be a single server (unless an error was previous raised during a
        reconciliation loop. ServerHandle can be any type that is sufficient for
        _remove_server_handle to remove the service."""

    @abstractmethod
    def _create_new_server_endpoint(
        self, location_name: str, metadata: CodeDeploymentMetadata
    ) -> GrpcServerEndpoint:
        """Create a new server for the given location using the given metadata as configuration
        and return a GrpcServerEndpoint indicating a hostname/port that can be used to access
        the server. Should result in an additional handle being returned from _get_server_handles_for_location."""

    @abstractmethod
    def _remove_server_handle(self, server_handle: ServerHandle) -> None:
        """Shut down any resources associated with the given handle. Called both during updates
        to spin down the old server once a new server has been spun up, and during removal."""

    @abstractmethod
    def _cleanup_servers(self):
        """Remove all servers, across all locations."""

    @abstractmethod
    def get_step_handler(self, execution_config: Optional[Dict]) -> StepHandler:
        pass

    @abstractmethod
    def run_launcher(self) -> WatchfulRunLauncher:
        pass

    def start(self):
        super().start()

        self._cleanup_servers()

        self._reconcile_grpc_metadata_thread = threading.Thread(
            target=self._reconcile_thread,
            args=(self._reconcile_grpc_metadata_shutdown_event,),
            name="grpc-reconcile-watch",
        )
        self._reconcile_grpc_metadata_thread.daemon = True
        self._reconcile_grpc_metadata_thread.start()

    def __exit__(self, exception_type, exception_value, traceback):
        if self._reconcile_grpc_metadata_thread:
            self._reconcile_grpc_metadata_shutdown_event.set()
            self._reconcile_grpc_metadata_thread.join()

        if self._started:
            self._cleanup_servers()

        super().__exit__(exception_value, exception_value, traceback)

    def update_grpc_metadata(
        self,
        desired_metadata: Dict[str, UserCodeLauncherEntry],
        upload_locations: Set[str],
    ):
        check.dict_param(
            desired_metadata,
            "desired_metadata",
            key_type=str,
            value_type=UserCodeLauncherEntry,
        )
        check.set_param(upload_locations, "upload_locations", str)

        with self._metadata_lock:
            self._desired_entries = desired_metadata
            self._upload_locations = self._upload_locations.union(upload_locations)

    def _get_repository_location_origin(
        self, location_name: str
    ) -> RegisteredRepositoryLocationOrigin:
        return RegisteredRepositoryLocationOrigin(location_name)

    def _reconcile_thread(self, shutdown_event):
        while True:
            shutdown_event.wait(USER_CODE_LAUNCHER_RECONCILE_INTERVAL)
            if shutdown_event.is_set():
                break

            try:
                self.reconcile()
            except Exception:
                self._logger.error(
                    "Failure updating user code servers: {exc_info}".format(
                        exc_info=sys.exc_info(),
                    )
                )

    def reconcile(self):
        with self._metadata_lock:
            if self._desired_entries == None:
                # Wait for the first time the desired metadata is set before reconciling
                return

            self._reconcile()
            self._reconcile_count += 1

            if self._reconcile_count == 1:
                with self._is_workspace_ready_lock:
                    self._is_workspace_ready = True

    def _check_for_image(self, metadata: CodeDeploymentMetadata):
        if self.requires_images and not metadata.image:
            raise Exception(
                "Your agent's configuration requires you to specify an image. "
                "Use the `--image` flag when specifying your location to tell the agent "
                "which image to use to load your code."
            )

        if (not self.requires_images) and metadata.image:
            raise Exception(
                "Your agent's configuration cannot load locations that specify a Docker "
                "image. Either update your location to not include an image, or change the `user_code_launcher` "
                "field in your agent's `dagster.yaml` file to a launcher that can load Docker images. "
            )

    def _reconcile(self):
        diff = diff_serializable_namedtuple_map(self._desired_entries, self._actual_entries)

        to_update_keys = diff.to_add.union(diff.to_update)
        update_gens: Dict[str, Iterator[GrpcServerEndpoint]] = {}
        for to_update_key in to_update_keys:
            try:

                code_deployment_metadata = self._desired_entries[
                    to_update_key
                ].code_deployment_metadata

                self._check_for_image(code_deployment_metadata)
                update_gens[to_update_key] = self._gen_update_server(
                    to_update_key,
                    code_deployment_metadata,
                )
                new_updated_endpoint: GrpcServerEndpoint = next(update_gens[to_update_key])
            except Exception:
                error_info = serializable_error_info_from_exc_info(sys.exc_info())
                self._logger.error(
                    "Error while updating server for {to_update_key}: {error_info}".format(
                        to_update_key=to_update_key,
                        error_info=error_info,
                    )
                )
                new_updated_endpoint: SerializableErrorInfo = error_info

            with self._grpc_endpoints_lock:
                self._grpc_endpoints[to_update_key] = new_updated_endpoint

            if to_update_key in self._upload_locations:
                self._upload_locations.remove(to_update_key)
                self._update_location_data(
                    to_update_key,
                    new_updated_endpoint,
                    self._desired_entries[to_update_key].code_deployment_metadata,
                )

        for to_update_key in to_update_keys:
            update_gen = update_gens.get(to_update_key)
            if update_gen:
                # Finish any remaining cleanup
                try:
                    list(update_gen)
                except Exception:
                    self._logger.error(
                        "Error while cleaning up after updating server for {to_update_key}: {error_info}".format(
                            to_update_key=to_update_key,
                            error_info=serializable_error_info_from_exc_info(sys.exc_info()),
                        )
                    )

            self._actual_entries[to_update_key] = self._desired_entries[to_update_key]

        for to_remove_key in diff.to_remove:
            try:
                self._remove_server(to_remove_key)
            except Exception:
                self._logger.error(
                    "Error while removing server for {to_remove_key}: {error_info}".format(
                        to_remove_key=to_remove_key,
                        error_info=serializable_error_info_from_exc_info(sys.exc_info()),
                    )
                )

            with self._grpc_endpoints_lock:
                del self._grpc_endpoints[to_remove_key]
            del self._actual_entries[to_remove_key]

        # Upload any locations that were requested to be uploaded, but weren't updated
        # as part of this reconciliation loop
        for location in self._upload_locations:
            with self._grpc_endpoints_lock:
                endpoint = self._grpc_endpoints.get(location)

            self._update_location_data(
                location,
                endpoint,
                self._actual_entries[location].code_deployment_metadata,
            )

        self._upload_locations = set()

    def get_grpc_endpoint(
        self, repository_location_origin: RepositoryLocationOrigin
    ) -> GrpcServerEndpoint:
        with self._grpc_endpoints_lock:
            location_name = repository_location_origin.location_name
            endpoint = self._grpc_endpoints.get(location_name)

        if not endpoint:
            raise Exception(f"No server endpoint exists for location {location_name}")

        if isinstance(endpoint, SerializableErrorInfo):
            # Consider raising the original exception here instead of a wrapped one
            raise Exception(
                f"Failure loading server endpoint for location {location_name}: {endpoint}"
            )

        return endpoint

    def get_grpc_endpoints(
        self,
    ) -> Dict[str, Union[GrpcServerEndpoint, SerializableErrorInfo]]:
        with self._grpc_endpoints_lock:
            return self._grpc_endpoints.copy()

    def _gen_update_server(
        self, location_name: str, new_metadata: CodeDeploymentMetadata
    ) -> Iterator[GrpcServerEndpoint]:
        # Update the server for the given location. Is a generator - should yield the new
        # GrpcServerEndpoint, then clean up any no longer needed resources
        existing_server_handles = self._get_server_handles_for_location(location_name)
        updated_server = self._create_new_server_endpoint(location_name, new_metadata)

        yield updated_server

        for server_handle in existing_server_handles:
            self._remove_server_handle(server_handle)

    def _remove_server(self, location_name: str):
        existing_server_handles = self._get_server_handles_for_location(location_name)
        for server_handle in existing_server_handles:
            self._remove_server_handle(server_handle)

    def _wait_for_server(
        self, host: str, port: int, timeout=15, socket: Optional[str] = None
    ) -> str:
        # Wait for the server to be ready (while also loading the server ID)
        server_id = None
        start_time = time.time()
        while True:
            client = DagsterGrpcClient(port=port, host=host, socket=socket)
            try:
                server_id = sync_get_server_id(client)
                break
            except Exception:
                pass

            if time.time() - start_time > timeout:
                raise Exception(f"Timed out waiting for server {host}:{port}")

            time.sleep(1)
        return server_id



from typing import Optional, Tuple

from dagster_cloud.execution.utils import TaskStatus


def check_on_container(client, container_name_or_id: str) -> Tuple[TaskStatus, Optional[str]]:
    try:
        container = client.containers.get(container_name_or_id)
    except Exception as e:
        return TaskStatus.NOT_FOUND, f"Hit Exception trying to get the container: {e}"

    if container.status == "running":
        return TaskStatus.RUNNING, None

    try:
        container_info = container.wait(timeout=0.1)
    except Exception as e:
        return (
            TaskStatus.UNKNOWN,
            f"Container status is {container.status}. Hit exception attempting to get its return code: {e}",
        )
    else:
        ret_code = container_info.get("StatusCode")
        if ret_code == 0:
            return TaskStatus.SUCCESS, None
        elif ret_code:
            return (
                TaskStatus.FAILURE,
                f"Container status is {container.status}. Return code is {str(ret_code)}.",
            )

    return TaskStatus.UNKNOWN, f"Container status is {container.status}. Return code is unknown."



from pathlib import Path

import requests
from typer import Argument, Typer

from ...cli import ui
from ...cli.config_utils import dagster_cloud_options
from ...errors import raise_http_error
from ...headers.impl import get_dagster_cloud_api_headers

app = Typer(help="Customize your SAML settings.")


@app.command(name="upload-identity-provider-metadata")
@dagster_cloud_options(allow_empty=True, requires_url=True)
def upload_identity_provider_metadata_command(
    url: str,
    api_token: str,
    metadata_file: Path = Argument(..., readable=True, metavar="METADATA_FILE_PATH"),
):
    """Uploads your identity provider's metadata to enable SAML Single sign-on (SSO)."""

    with open(metadata_file, "r") as f:
        response = requests.post(
            url=f"{url}/upload_idp_metadata",
            headers=get_dagster_cloud_api_headers(api_token),
            files={"metadata.xml": f},
        )

    raise_http_error(response)

    ui.print("The identity provider metadata was successfully uploaded.")



import collections
import functools
import inspect
import os
from typing import List, NamedTuple, Optional

import yaml
from click import Context
from dagster_cloud.errors import DagsterCloudHTTPError
from typer import Option

from . import gql, ui
from .utils import add_options

DEPLOYMENT_CLI_ARGUMENT = "deployment"
DEPLOYMENT_ENV_VAR_NAME = "DAGSTER_CLOUD_DEPLOYMENT"

ORGANIZATION_CLI_ARGUMENT = "organization"
ORGANIZATION_ENV_VAR_NAME = "DAGSTER_CLOUD_ORGANIZATION"

TOKEN_CLI_ARGUMENT = "api-token"
TOKEN_CLI_ARGUMENT_VAR = TOKEN_CLI_ARGUMENT.replace("-", "_")
TOKEN_ENV_VAR_NAME = "DAGSTER_CLOUD_API_TOKEN"

URL_CLI_ARGUMENT = "url"
URL_ENV_VAR_NAME = "DAGSTER_CLOUD_DAGIT_URL"

DEFAULT_CLOUD_CLI_FOLDER = os.path.join(os.path.expanduser("~"), ".dagster_cloud_cli")
DEFAULT_CLOUD_CLI_CONFIG = os.path.join(DEFAULT_CLOUD_CLI_FOLDER, "config")


class DagsterCloudCliConfig(
    NamedTuple(
        "_DagsterCloudCliConfig",
        [
            ("organization", Optional[str]),
            ("default_deployment", Optional[str]),
            ("user_token", Optional[str]),
        ],
    )
):
    __slots__ = ()

    def __new__(cls, **kwargs):
        none_defaults = {k: (kwargs[k] if k in kwargs else None) for k in cls._fields}
        return super(DagsterCloudCliConfig, cls).__new__(cls, **none_defaults)


def get_config_path():
    return os.getenv("DAGSTER_CLOUD_CLI_CONFIG", DEFAULT_CLOUD_CLI_CONFIG)


def write_config(config: DagsterCloudCliConfig):
    """
    Writes the given config object to the CLI config file.
    """
    config_path = get_config_path()
    os.makedirs(os.path.dirname(config_path), exist_ok=True)
    with open(config_path, "w") as f:
        config_dict = {k: v for k, v in config._asdict().items() if v is not None}
        f.write(yaml.dump(config_dict))


def read_config() -> DagsterCloudCliConfig:
    """
    Reads the CLI config file into a config object.
    """
    config_path = get_config_path()
    if not os.path.isfile(config_path):
        return DagsterCloudCliConfig()

    with open(config_path, "r") as f:
        raw_in = yaml.load(f.read(), Loader=yaml.SafeLoader)
        return DagsterCloudCliConfig(**raw_in)


def get_deployment(ctx: Optional[Context] = None) -> Optional[str]:
    """
    Gets the configured deployment to target.
    Highest precedence is a deployment argument, then `DAGSTER_CLOUD_DEPLOYMENT`
    env var, then `~/.dagster_cloud_cli/config` default.
    """
    if ctx and ctx.params.get(DEPLOYMENT_CLI_ARGUMENT):
        return ctx.params[DEPLOYMENT_CLI_ARGUMENT]
    return os.getenv(DEPLOYMENT_ENV_VAR_NAME, read_config().default_deployment)


def get_organization(ctx: Optional[Context] = None) -> Optional[str]:
    """
    Gets the configured organization to target.
    Highest precedence is an organization argument, then `DAGSTER_CLOUD_ORGANIZATION`
    env var, then `~/.dagster_cloud_cli/config` value.
    """
    if ctx and ctx.params.get(ORGANIZATION_CLI_ARGUMENT):
        return ctx.params[ORGANIZATION_CLI_ARGUMENT]
    return os.getenv(ORGANIZATION_ENV_VAR_NAME, read_config().organization)


def get_user_token(ctx: Optional[Context] = None) -> Optional[str]:
    """
    Gets the configured user token to use.
    Highest precedence is an api-token argument, then `DAGSTER_CLOUD_API_TOKEN`
    env var, then `~/.dagster_cloud_cli/config` value.
    """
    if ctx and ctx.params.get(TOKEN_CLI_ARGUMENT_VAR):
        return ctx.params[TOKEN_CLI_ARGUMENT_VAR]
    return os.getenv(TOKEN_ENV_VAR_NAME, read_config().user_token)


def available_deployment_names(ctx, incomplete: str = "") -> List[str]:
    """
    Gets a list of deployment names given the Typer Context, used for CLI completion.
    """
    organization = get_organization(ctx=ctx)
    user_token = get_user_token(ctx=ctx)

    if not organization or not user_token:
        return []
    try:
        deployments = gql.fetch_deployments(
            gql.graphql_client_from_url(gql.url_from_config(organization), user_token)
        )
        names = [deployment["deploymentName"] for deployment in deployments]
        return [name for name in names if name.startswith(incomplete)]
    except DagsterCloudHTTPError:
        return []


def get_url(ctx: Optional[Context] = None) -> Optional[str]:
    """
    Gets the url passed in or from the environment.
    """
    if ctx and ctx.params.get(URL_CLI_ARGUMENT):
        return ctx.params[URL_CLI_ARGUMENT]
    return os.getenv(URL_ENV_VAR_NAME)


# Typer Option definitions for common CLI config options (organization, deployment, user token)
ORGANIZATION_OPTION = Option(
    get_organization, "--organization", "-o", help="Organization to target."
)
DEPLOYMENT_OPTION = Option(
    get_deployment,
    "--deployment",
    "-d",
    help="Deployment to target.",
    autocompletion=available_deployment_names,
)
USER_TOKEN_OPTION = Option(
    get_user_token,
    "--api-token",
    "--user-token",
    "-u",
    help="Cloud user token.",
    show_default=False,
)
URL_OPTION = Option(
    get_url,
    "--url",
    help="[DEPRECATED] Your Dagster Cloud url, in the form of 'https://{ORGANIZATION_NAME}.dagster.cloud/{DEPLOYMENT_NAME}'.",
)


def dagster_cloud_options(
    allow_empty: bool = False, allow_empty_deployment: bool = False, requires_url: bool = False
):
    """
    Apply this decorator to Typer commands to make them take the
    `organization`, `deployment`, and `api-token` arguments.
    These values are passed as keyword arguments.

    Unless `allow_empty` or `allow_empty_deployment` are set, an error
    will be raised if these values are not specified as arguments or via config/env var.

    Set `requires_url` if the command needs a Dagit url; if so, this will be provided
    via the `url` keyword argument.
    """

    def decorator(to_wrap):
        wrapped_sig = inspect.signature(to_wrap)
        params = collections.OrderedDict(wrapped_sig.parameters)

        options = {
            ORGANIZATION_CLI_ARGUMENT: (str, ORGANIZATION_OPTION),
            TOKEN_CLI_ARGUMENT_VAR: (str, USER_TOKEN_OPTION),
        }

        # For requests that need a graphql client, a user can supply a URL (though this
        # is a hidden option), or an organization and deployment
        if requires_url:
            options[URL_CLI_ARGUMENT] = (str, URL_OPTION)

        has_deployment_param = (
            DEPLOYMENT_CLI_ARGUMENT in params
            and params[DEPLOYMENT_CLI_ARGUMENT].default is params[DEPLOYMENT_CLI_ARGUMENT].empty
        ) or requires_url
        if has_deployment_param:
            options[DEPLOYMENT_CLI_ARGUMENT] = (str, DEPLOYMENT_OPTION)

        with_options = add_options(options)(to_wrap)

        @functools.wraps(with_options)
        def wrap_function(*args, **kwargs):

            # If underlying command needs a URL and none is explicitly provided,
            # generate one from the organization + deployment
            if (
                requires_url
                and not kwargs.get(URL_CLI_ARGUMENT)
                and kwargs.get(ORGANIZATION_CLI_ARGUMENT)
                and kwargs.get(DEPLOYMENT_CLI_ARGUMENT)
            ):
                kwargs[URL_CLI_ARGUMENT] = gql.url_from_config(
                    organization=kwargs.get(ORGANIZATION_CLI_ARGUMENT),
                    deployment=kwargs.get(DEPLOYMENT_CLI_ARGUMENT),
                )

            lacking_url = requires_url and not kwargs.get(URL_CLI_ARGUMENT)

            # Raise errors if important options are not provided
            if not kwargs.get(TOKEN_CLI_ARGUMENT_VAR) and (not allow_empty or lacking_url):
                raise ui.error(
                    f"A Dagster Cloud user token must be specified for this command.\n\nYou may specify a token by:\n"
                    f"- Providing the {ui.as_code('--' + TOKEN_CLI_ARGUMENT)} parameter.\n"
                    f"- Setting the {ui.as_code(TOKEN_ENV_VAR_NAME)} environment variable.\n"
                    f"- Specifying {ui.as_code('user_token')} in your config file ({get_config_path()}), run {ui.as_code('dagster-cloud configure')}."
                )
            if not kwargs.get(ORGANIZATION_CLI_ARGUMENT) and (not allow_empty or lacking_url):
                raise ui.error(
                    f"A Dagster Cloud organization must be specified for this command.\n\nYou may specify your organization by:\n"
                    f"- Providing the {ui.as_code('--' + ORGANIZATION_CLI_ARGUMENT)} parameter.\n"
                    f"- Setting the {ui.as_code(ORGANIZATION_ENV_VAR_NAME)} environment variable.\n"
                    f"- Specifying {ui.as_code('organization')} in your config file ({get_config_path()}), run {ui.as_code('dagster-cloud configure')}."
                )
            if (
                DEPLOYMENT_CLI_ARGUMENT in kwargs
                and not kwargs.get(DEPLOYMENT_CLI_ARGUMENT)
                and (not (allow_empty or allow_empty_deployment) or lacking_url)
            ):
                raise ui.error(
                    f"A Dagster Cloud deployment must be specified for this command.\n\nYou may specify a deployment by:\n"
                    f"- Providing the {ui.as_code('--' + DEPLOYMENT_CLI_ARGUMENT)} parameter.\n"
                    f"- Setting the {ui.as_code(DEPLOYMENT_ENV_VAR_NAME)} environment variable.\n"
                    f"- Running {ui.as_code('dagster-cloud config set-deployment <deployment_name>')}.\n"
                    f"- Specifying {ui.as_code('default_deployment')} in your config file ({get_config_path()})."
                )

            new_kwargs = dict(kwargs)
            with_options(*args, **new_kwargs)

        return wrap_function

    return decorator



# pylint: disable=unused-argument
import mock
from dagster_cloud.agent.cli import app
from typer.testing import CliRunner
import os
import tempfile
from dagster.core.test_utils import environ
from pathlib import Path


def _check_dagster_home(expected: str):
    def _check_inner():
        assert os.getenv("DAGSTER_HOME") == expected

    return _check_inner


def test_run_command_with_env():
    runner = CliRunner()

    with tempfile.TemporaryDirectory() as temp_dir:
        with environ({"DAGSTER_HOME": temp_dir}):
            with mock.patch(
                "dagster_cloud.agent.cli.run_local_agent", _check_dagster_home(temp_dir)
            ):
                result = runner.invoke(app, [])

                assert result.exit_code == 0, result.output + " : " + str(result.exception)


def test_run_command_with_no_home():
    runner = CliRunner()

    with environ({"DAGSTER_HOME": None}):
        result = runner.invoke(app, [])

        assert result.exit_code == 1, result.output + " : " + str(result.exception)
        assert "No directory provided" in result.output


def test_run_command_with_argument():
    runner = CliRunner()

    # Test param works
    with tempfile.TemporaryDirectory() as temp_dir:
        with mock.patch(
            "dagster_cloud.agent.cli.run_local_agent",
            _check_dagster_home(str(Path(temp_dir).resolve())),
        ):
            result = runner.invoke(app, [temp_dir])

            assert result.exit_code == 0, result.output + " : " + str(result.exception)

    # Test param overrides DAGSTER_HOME env var
    with tempfile.TemporaryDirectory() as temp_dir, tempfile.TemporaryDirectory() as temp_dir_env:
        with environ({"DAGSTER_HOME": temp_dir_env}):
            with mock.patch(
                "dagster_cloud.agent.cli.run_local_agent",
                _check_dagster_home(str(Path(temp_dir).resolve())),
            ):
                result = runner.invoke(app, [temp_dir])

                assert result.exit_code == 0, result.output + " : " + str(result.exception)



from collections import namedtuple
from typing import Any, Dict, List

from dagster import Field, check
from dagster.config.source import BoolSourceType, IntSourceType, StringSourceType
from dagster.serdes.utils import create_snapshot_id


class SerializableNamedtupleMapDiff(
    namedtuple(
        "_SerializableNamedtupleMapDiff",
        "to_add to_update to_remove",
    )
):
    def __new__(
        cls,
        to_add,
        to_update,
        to_remove,
    ):
        return super(SerializableNamedtupleMapDiff, cls).__new__(
            cls,
            check.set_param(to_add, "to_add", str),
            check.set_param(to_update, "to_update", str),
            check.set_param(to_remove, "to_remove", str),
        )


def diff_serializable_namedtuple_map(desired_map, actual_map, force_update_keys=None):
    desired_keys = set(desired_map.keys())
    actual_keys = set(actual_map.keys())
    force_update_keys = check.opt_set_param(force_update_keys, "force_update_keys", str)

    to_add = desired_keys.difference(actual_keys)
    to_remove = actual_keys.difference(desired_keys)

    existing = actual_keys.intersection(desired_keys)

    to_update = {
        existing_key
        for existing_key in existing
        if create_snapshot_id(desired_map[existing_key])
        != create_snapshot_id(actual_map[existing_key])
        or existing_key in force_update_keys
    }

    return SerializableNamedtupleMapDiff(to_add, to_update, to_remove)


def get_env_names_from_config(
    config_schema: Dict[str, Field], config_dict: Dict[str, Any]
) -> List[str]:
    env_vars = []
    for field_name, field in config_schema.items():
        config_type = field.config_type
        if isinstance(
            config_type, (StringSourceType, IntSourceType, BoolSourceType)
        ) and isinstance(config_dict.get(field_name), dict):
            env_name = config_dict[field_name].get("env")
            if env_name:
                env_vars.append(env_name)

    return env_vars



import os
import uuid
from pathlib import Path
from typing import Optional

from dagster.utils.interrupts import capture_interrupts
from dagster_cloud.agent.dagster_cloud_agent import DagsterCloudAgent
from dagster_cloud.instance import DagsterCloudAgentInstance
from typer import Argument, Typer
from ...cli import ui
from dagster.core.errors import DagsterHomeNotSetError

app = Typer()


def agent_home_exception():
    dagster_home_loc = (
        (
            f"No Dagster config provided in specified directory {os.getenv('DAGSTER_HOME')}. "
            "You must specify the location of a directory containing a dagster.yaml "
            "file as a parameter or by setting the DAGSTER_HOME environment variable."
        )
        if os.getenv("DAGSTER_HOME")
        else (
            "No directory provided or DAGSTER_CLOUD environment variable set. "
            "You must supply the location of a directory containing a dagster.yaml "
            "file as a parameter or by setting the DAGSTER_HOME environment variable."
        )
    )
    return ui.error(f"No Dagster config found.\n\n{dagster_home_loc}")


def run_local_agent():
    try:
        with DagsterCloudAgentInstance.get() as instance:
            user_code_launcher = instance.user_code_launcher
            user_code_launcher.start()

            with DagsterCloudAgent() as agent:
                agent.run_loop(instance, user_code_launcher, agent_uuid=str(uuid.uuid4()))
    except DagsterHomeNotSetError:
        raise agent_home_exception()


def run_local_agent_in_environment(dagster_home: Optional[Path]):
    with capture_interrupts():
        try:
            old_env = dict(os.environ)
            if dagster_home:
                os.environ["DAGSTER_HOME"] = str(dagster_home.resolve())
            run_local_agent()
        finally:
            os.environ.clear()
            os.environ.update(old_env)


@app.command(
    help=(
        "Runs the Dagster Cloud agent. Sources config from a dagster.yaml located in the "
        "optional provided directory, otherwise uses the directory specified by the DAGSTER_HOME "
        "environment variable."
    )
)
def run(dagster_home: Optional[Path] = Argument(None)):
    run_local_agent_in_environment(dagster_home)



from dagster_cloud.agent.cli import app as agent_app
from dagster_cloud.settings import app as settings_app
from dagster_cloud.workspace.cli import app as workspace_app
from typer import Typer

from .commands.config import app as config_app
from .commands.config import app_configure as configure_app


def _import_commands(parent: Typer, child: Typer) -> None:
    """
    Copies the commands from one Typer app to another.
    Equivalent of `add_typer` but doesn't add a subcommand.
    """
    for command in child.registered_commands:
        parent.registered_commands.append(command)


app = Typer(help="CLI tools for working with Dagster Cloud.")
app.add_typer(agent_app, name="agent")
app.add_typer(workspace_app, name="workspace")
app.add_typer(settings_app, name="settings")
app.add_typer(config_app, name="config")
_import_commands(app, configure_app)

if __name__ == "__main__":
    app()



from collections import namedtuple

from dagster import check
from dagster.serdes import whitelist_for_serdes


@whitelist_for_serdes
class CodeDeploymentMetadata(
    namedtuple("_CodeDeploymentMetadata", "image python_file package_name")
):
    def __new__(cls, image=None, python_file=None, package_name=None):
        check.invariant(
            (python_file or package_name) and not (python_file and package_name),
            "Must supply exactly one of a file name or a package name",
        )
        return super(CodeDeploymentMetadata, cls).__new__(
            cls,
            check.opt_str_param(image, "image"),
            check.opt_str_param(python_file, "python_file"),
            check.opt_str_param(package_name, "package_name"),
        )



from dagster_cloud.execution.watchful_run_launcher.base import WatchfulRunLauncher
from dagster_k8s import K8sRunLauncher
from dagster_k8s.job import get_job_name_from_run_id

from . import WatchfulRunLauncher


class WatchfulK8sRunLauncher(K8sRunLauncher, WatchfulRunLauncher):
    def check_run_health(self, run_id):
        job_name = get_job_name_from_run_id(run_id)
        job = self._batch_api.read_namespaced_job(namespace=self.job_namespace, name=job_name)
        pipeline_run = self._instance.get_run_by_id(run_id)
        if job.status.failed:
            self._instance.report_run_failed(
                pipeline_run, f"Kubernetes job {job_name} failed in your cluster."
            )



from contextlib import contextmanager

import pytest
from dagster.core.test_utils import instance_for_test
from dagster_postgres.utils import get_conn_string, wait_for_connection
from ursula.test_utils import make_live_graphql_client


@pytest.fixture(name="agent_instance")
def agent_instance_fixture(agent_token, ursula_graphql_client):  # pylint: disable=unused-argument
    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.user_code_launcher",
                "class": "ProcessUserCodeLauncher",
            },
            "dagster_cloud_api": {
                "url": "http://localhost:2873",
                "agent_token": agent_token,
            },
            "compute_logs": {
                "module": "dagster.core.storage.noop_compute_log_manager",
                "class": "NoOpComputeLogManager",
            },
        }
    ) as instance:
        yield instance


@pytest.fixture(name="agent_replicas_instance")
def agent_replicas_instance_fixture(
    agent_token, ursula_graphql_client
):  # pylint: disable=unused-argument
    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.user_code_launcher",
                "class": "ProcessUserCodeLauncher",
            },
            "dagster_cloud_api": {
                "url": "http://localhost:2873",
                "agent_token": agent_token,
            },
            "compute_logs": {
                "module": "dagster.core.storage.noop_compute_log_manager",
                "class": "NoOpComputeLogManager",
            },
            "agent_replicas": {"enabled": True},
        }
    ) as instance:
        yield instance


@pytest.fixture(name="host_storage_hostnames", scope="module")
def host_storage_hostnames_fixture(docker_compose_cm):
    with docker_compose_cm() as docker_compose:
        host_storage_hostnames = docker_compose
        postgres_hostname = host_storage_hostnames["postgres"]

        wait_for_connection(
            get_conn_string(
                username="test",
                password="test",
                hostname=postgres_hostname,
                db_name="test",
            ),
            retry_limit=10,
            retry_wait=3,
        )

        yield host_storage_hostnames


@pytest.fixture(name="host_instance", scope="module")
def host_instance_fixture(host_storage_hostnames, agent_token):
    with _create_host_instance(host_storage_hostnames) as instance:
        instance.wipe_db()
        assert instance.get_alembic_rev() is None

        instance.init_db()
        assert instance.get_alembic_rev() != None

        organization_id = instance.cloud_storage.create_organization("acme")
        instance.cloud_storage.create_deployment(organization_id, "sandbox")

        test_user = instance.cloud_storage.create_or_update_user(email="test@acme.com")

        instance.cloud_storage.create_agent_token(
            organization_id,
            test_user.user_id,
            token_value=agent_token,
        )

        instance.check_loaded()
        with instance.for_deployment("acme", "sandbox") as scoped_instance:
            yield scoped_instance
        instance.wipe_db()
        assert instance.get_alembic_rev() is None


@contextmanager
def _create_host_instance(
    host_storage_hostnames,
    set_dagster_home=True,
):
    postgres_hostname = host_storage_hostnames["postgres"]
    redis_hostname = host_storage_hostnames["redis"]
    postgres_config = {
        "postgres_db": {
            "hostname": postgres_hostname,
            "username": "test",
            "password": "test",
            "db_name": "test",
        },
    }
    instance_overrides = {
        "run_storage": {
            "module": "ursula.storage.host_cloud.run_storage",
            "class": "PostgresCloudRunStorage",
            "config": postgres_config,
        },
        "event_log_storage": {
            "module": "ursula.storage.host_cloud.event_log_storage",
            "class": "PostgresCloudEventLogStorage",
            "config": postgres_config,
        },
        "schedule_storage": {
            "module": "ursula.storage.host_cloud.schedule_storage",
            "class": "PostgresCloudScheduleStorage",
            "config": postgres_config,
        },
        "instance_class": {
            "module": "ursula.instance",
            "class": "UnscopedHostInstance",
            "config": {},
        },
        "cloud_storage": {
            "module": "ursula.storage.host_cloud.cloud_storage",
            "class": "PostgresCloudStorage",
            "config": postgres_config,
        },
        "monitoring_storage": {
            "module": "ursula.storage.host_cloud.monitoring_storage",
            "class": "PostgresMonitoringStorage",
            "config": postgres_config,
        },
        "email_client": {
            "module": "ursula.daemon.monitoring.email.recording",
            "class": "RecordingEmailClient",
            "config": {},
        },
        "redis": {
            "host": redis_hostname,
        },
    }

    with instance_for_test(
        overrides=instance_overrides, set_dagster_home=set_dagster_home
    ) as instance:
        yield instance


@pytest.fixture(name="ursula_graphql_client", scope="module")
def ursula_graphql_client_fixture(agent_token, host_instance):  # pylint: disable=unused-argument
    yield from make_live_graphql_client(agent_token=agent_token)



from dagster_cloud.execution.utils import TaskStatus
from dagster_cloud.execution.utils.docker import check_on_container
from dagster_cloud.execution.watchful_run_launcher.base import WatchfulRunLauncher
from dagster_docker import DockerRunLauncher
from dagster_docker.docker_run_launcher import DOCKER_CONTAINER_ID_TAG


class WatchfulDockerRunLauncher(DockerRunLauncher, WatchfulRunLauncher):
    def check_run_health(self, run_id):
        run = self._instance.get_run_by_id(run_id)
        container_id = run.tags.get(DOCKER_CONTAINER_ID_TAG)
        status, msg = check_on_container(self._get_client(), container_id)

        if status not in [TaskStatus.RUNNING, TaskStatus.SUCCESS]:
            self._instance.report_run_failed(run, f"Container {container_id} is not running: {msg}")



from typing import List

from typer import Option, Typer

from ...cli import gql, ui
from ...cli.config_utils import dagster_cloud_options

app = Typer(help="Customize your login settings.")


SET_ALLOWLISTED_DOMAINS_MUTATION = """
    mutation ($allowlistedDomains: [String!]) {
        setAllowlistedDomains(allowlistedDomains: $allowlistedDomains) {
            __typename
            ... on DagsterCloudOrganization {
                name
                allowlistedDomains
            }
            ... on PythonError {
                message
                stack
            }
        }
    }
    """


@app.command(name="set-allowlisted-domains")
@dagster_cloud_options(allow_empty=True, requires_url=True)
def set_allowlisted_domains_command(
    url: str,
    api_token: str,
    domains: List[str] = Option(..., "--domain", "-d"),
):
    """
    Configures an organization's allowed domains.

    After running this command, only new users with email addresses from the allowed domains will be
    able to authenticate into their organization's Dagster Cloud instance.

    If a user does not already have a Dagster Cloud account, but their email matches one of the
    allowed domains, an account is created on their behalf.
    """
    client = gql.graphql_client_from_url(url, api_token)
    response = client.execute(
        SET_ALLOWLISTED_DOMAINS_MUTATION,
        variable_values={"allowlistedDomains": domains},
    )

    if response["data"]["setAllowlistedDomains"]["__typename"] == "DagsterCloudOrganization":
        ui.print(
            "Organization '{organization_name}' now has its allowlisted domains set to {domains}. "
            "Only users with email addresses from the allowed domains will be able to log into their "
            "organization's Dagster Cloud instance.".format(
                organization_name=response["data"]["setAllowlistedDomains"]["name"],
                domains=domains,
            )
        )
    else:
        raise ui.error(f"Unable to add domains to allowlist: {response}")


@app.command(name="clear-allowlisted-domains")
@dagster_cloud_options(allow_empty=True, requires_url=True)
def clear_allowlisted_domains_command(
    url: str,
    api_token: str,
):
    """
    Removes all the allowed domains from an organization.

    After running this command, only users with email addresses corresponding to an existing
    Dagster Cloud account will be able to authenticate into their organization's
    Dagster Cloud instance.
    """
    client = gql.graphql_client_from_url(url, api_token)
    response = client.execute(
        SET_ALLOWLISTED_DOMAINS_MUTATION,
        variable_values={"allowlistedDomains": []},
    )

    if response["data"]["setAllowlistedDomains"]["__typename"] == "DagsterCloudOrganization":
        ui.print(
            "Organization '{organization_name}' now has its allowlisted domains set to {domains}. "
        )
    else:
        raise ui.error(f"Unable to add domains to allowlist: {response}")



import uuid

import boto3
import botocore

from .service import Service


class Client:
    def __init__(
        self,
        cluster_name,
        subnet_ids,
        service_discovery_namespace_id,
        log_group,
        execution_role_arn,
    ):
        self.ecs = boto3.client("ecs")
        self.logs = boto3.client("logs")
        self.service_discovery = boto3.client("servicediscovery")
        self.ec2 = boto3.resource("ec2")

        self.cluster_name = cluster_name.split("/")[-1]
        self.subnet_ids = subnet_ids
        self.service_discovery_namespace_id = service_discovery_namespace_id
        self.log_group = log_group
        self.execution_role_arn = execution_role_arn

    @property
    def namespace(self):
        namespace = (
            self.service_discovery.get_namespace(
                Id=self.service_discovery_namespace_id,
            )
            .get("Namespace")
            .get("Name")
        )
        return namespace

    @property
    def taggable(self):
        settings = self.ecs.list_account_settings(
            name="serviceLongArnFormat",
            effectiveSettings=True,
        )
        return settings["settings"][0]["value"] == "enabled"

    def register_task_definition(self, name, image, command, task_role_arn=None, env=None):
        if not env:
            env = {}

        kwargs = dict(
            family=name,
            requiresCompatibilities=["FARGATE"],
            networkMode="awsvpc",
            containerDefinitions=[
                {
                    "name": name,
                    "image": image,
                    "environment": [{"name": key, "value": value} for key, value in env.items()],
                    "command": command,
                    "logConfiguration": {
                        "logDriver": "awslogs",
                        "options": {
                            "awslogs-group": self.log_group,
                            "awslogs-region": self.ecs.meta.region_name,
                            "awslogs-stream-prefix": name,
                        },
                    },
                },
            ],
            executionRoleArn=self.execution_role_arn,
            # https://docs.aws.amazon.com/AmazonECS/latest/developerguide/task-cpu-memory-error.html
            cpu="256",
            memory="512",
        )

        if task_role_arn:
            kwargs.update(dict(taskRoleArn=task_role_arn))

        task_definition_arn = (
            self.ecs.register_task_definition(**kwargs)
            .get("taskDefinition")
            .get("taskDefinitionArn")
        )

        return task_definition_arn

    def create_service(
        self,
        name,
        image,
        command,
        task_role_arn=None,
        env=None,
        tags=None,
        register_service_discovery=True,
    ):
        # Append a unique suffix to the service name so we can do blue/green
        # deploys without setting up a load balancer.
        service_name = f"{name}_{uuid.uuid4().hex[:7]}"

        # Register a task definition
        task_definition_arn = self.register_task_definition(
            name=name,
            image=image,
            task_role_arn=task_role_arn,
            command=command,
            env=env,
        )

        service_registry_arn = None
        # Configure service discovery
        if register_service_discovery:
            service_registry_arn = self._create_service_registry(
                service_name=service_name,
                tags=tags,
            )

        # Create the service
        service = self._create_service(
            service_name=service_name,
            service_registry_arn=service_registry_arn,
            task_definition_arn=task_definition_arn,
            tags=tags,
        )

        # Poll until the service is available
        self._wait_for_service(service)

        return service

    def delete_service(self, service):
        # Reduce running tasks to 0
        self.ecs.update_service(
            cluster=self.cluster_name,
            service=service.name,
            desiredCount=0,
        )
        self.ecs.get_waiter("services_stable").wait(
            cluster=self.cluster_name,
            services=[service.name],
        )

        # Delete the ECS service
        self.ecs.delete_service(
            cluster=self.cluster_name,
            service=service.name,
        )
        self.ecs.get_waiter("services_inactive").wait(
            cluster=self.cluster_name,
            services=[service.name],
        )

        # Delete service discovery
        service_discovery_id = self._get_service_discovery_id(
            service.hostname,
        )
        if service_discovery_id:
            self.service_discovery.delete_service(
                Id=service_discovery_id,
            )

    def list_services(self):
        services = []

        paginator = self.ecs.get_paginator("list_services")
        for page in paginator.paginate(cluster=self.cluster_name):
            for arn in page.get("serviceArns"):
                service = Service(client=self, arn=arn)
                services.append(service)

        return services

    def run_task(self, name, image, command):
        task_definition_arn = self.register_task_definition(
            name=name,
            image=image,
            command=command,
        )

        task_arn = (
            self.ecs.run_task(
                taskDefinition=task_definition_arn,
                cluster=self.cluster_name,
                launchType="FARGATE",
                networkConfiguration={
                    "awsvpcConfiguration": {
                        "subnets": self.subnet_ids,
                        "assignPublicIp": self._assign_public_ip(),
                    },
                },
            )
            .get("tasks", [{}])[0]
            .get("taskArn")
        )

        self.ecs.get_waiter("tasks_stopped").wait(
            cluster=self.cluster_name,
            tasks=[task_arn],
        )

        exit_code = (
            self.ecs.describe_tasks(
                cluster=self.cluster_name,
                tasks=[task_arn],
            )
            .get("tasks", [{}])[0]
            .get("containers", [{}])[0]
            .get("exitCode")
        )

        if exit_code:
            task_id = task_arn.split("/")[-1]

            log_stream = f"{name}/{name}/{task_id}"

            events = self.logs.get_log_events(
                logGroupName=self.log_group,
                logStreamName=log_stream,
            ).get("events")

            raise Exception([event.get("message") for event in events])

        return True

    def _create_service_registry(self, service_name, tags=None):
        if not tags:
            tags = {}

        # https://docs.aws.amazon.com/AmazonECS/latest/developerguide/create-service-discovery.html
        service_registry_arn = (
            self.service_discovery.create_service(
                Name=service_name,
                NamespaceId=self.service_discovery_namespace_id,
                DnsConfig={
                    "DnsRecords": [
                        {"Type": "A", "TTL": 60},
                    ]
                },
                Tags=[{"Key": key, "Value": value} for key, value in tags.items()],
            )
            .get("Service", {})
            .get("Arn")
        )
        return service_registry_arn

    def _create_service(self, service_name, task_definition_arn, service_registry_arn, tags=None):
        params = dict(
            cluster=self.cluster_name,
            serviceName=service_name,
            taskDefinition=task_definition_arn,
            launchType="FARGATE",
            desiredCount=1,
            networkConfiguration={
                "awsvpcConfiguration": {
                    "subnets": self.subnet_ids,
                    "assignPublicIp": self._assign_public_ip(),
                },
            },
        )

        if service_registry_arn:
            params["serviceRegistries"] = [{"registryArn": service_registry_arn}]

        if tags and self.taggable:
            params["tags"] = [{"key": key, "value": value} for key, value in tags.items()]

        arn = self.ecs.create_service(**params).get("service").get("serviceArn")

        return Service(client=self, arn=arn)

    def _wait_for_service(self, service):
        service_name = service.name
        # Check if we can place a task; this is fairly quick
        # TODO: handle IAM eventual consistency; sometimes the first event will
        # read "ECS was unable to assume the role" but will resolve itself with
        # enough time:
        # https://docs.aws.amazon.com/IAM/latest/UserGuide/troubleshoot_general.html#troubleshoot_general_eventual-consistency
        try:
            self.ecs.get_waiter("services_stable").wait(
                cluster=self.cluster_name,
                services=[service_name],
                # 5 minutes
                WaiterConfig={
                    "Delay": 2,
                    "MaxAttempts": 150,
                },
            )

        except botocore.exceptions.WaiterError:
            service = self.ecs.describe_services(
                cluster=self.cluster_name,
                services=[service_name],
            ).get("services")[0]

            messages = [event.get("message") for event in service.get("events")]

            # These strings in event messages indicate we couldn't create a task
            # https://docs.aws.amazon.com/AmazonECS/latest/developerguide/service-event-messages.html#service-event-messages-list
            failures = ["unable", "unhealthy", "throttled"]
            for message in messages:
                if any(failure in message for failure in failures):
                    # TODO: Custom exception
                    raise Exception(messages)

        # Check if a task can start; this is pretty slow but
        # will return very quickly on the happy path
        # TODO: Can we make it fail faster if it can't pull the image?
        try:
            self.ecs.get_waiter("services_stable").wait(
                cluster=self.cluster_name,
                services=[service_name],
                # 5 minutes
                WaiterConfig={
                    "Delay": 1,
                    "MaxAttempts": 300,
                },
            )

        except botocore.exceptions.WaiterError:
            task_arns = self.ecs.list_tasks(
                cluster=self.cluster_name,
                serviceName=service_name,
                desiredStatus="STOPPED",
            ).get("taskArns")
            tasks = self.ecs.describe_tasks(cluster=self.cluster_name, tasks=task_arns).get("tasks")
            reasons = [task.get("stoppedReason") for task in tasks]
            raise Exception(reasons)

    def _get_service_discovery_id(self, hostname):
        service_name = hostname.split("." + self.namespace)[0]

        paginator = self.service_discovery.get_paginator("list_services")
        for page in paginator.paginate(
            Filters=[
                {
                    "Name": "NAMESPACE_ID",
                    "Values": [
                        self.service_discovery_namespace_id,
                    ],
                    "Condition": "EQ",
                },
            ],
        ):
            for service in page["Services"]:
                if service["Name"] == service_name:
                    return service["Id"]

    def _assign_public_ip(self):
        # https://docs.aws.amazon.com/AmazonECS/latest/userguide/fargate-task-networking.html
        # Assign a public IP if any of the subnets are public
        route_tables = self.ec2.route_tables.filter(
            Filters=[
                {"Name": "association.subnet-id", "Values": self.subnet_ids},
            ]
        )

        # Consider a subnet to be public if it has a route that targets
        # an internet gateway; private subnets have routes that target NAT gateways
        for route_table in route_tables:
            if any(route.nat_gateway_id for route in route_table.routes):
                return "DISABLED"
        return "ENABLED"



from typer import Argument, Context, Option, Typer

from .. import gql, ui
from typing import Optional
from ..config_utils import (
    DagsterCloudCliConfig,
    available_deployment_names,
    dagster_cloud_options,
    read_config,
    write_config,
)

app = Typer()


@app.command()
@dagster_cloud_options()
def set_deployment(
    ctx: Context, deployment: str = Argument(..., autocompletion=available_deployment_names)
):
    deployments = available_deployment_names(ctx=ctx)
    if not deployment or not deployment in deployments:
        raise ui.error(f"Deployment {ui.as_code(deployment)} not found")

    config = read_config()
    new_config = config._replace(default_deployment=deployment)
    write_config(new_config)

    ui.print(f"Default deployment changed to {ui.as_code(deployment)}")


@app.command()
def view(
    show_token: bool = Option(
        False, "--show-token", "-s", help="Whether to display the user token in plaintext."
    )
):
    config = read_config()
    if not show_token and config.user_token:
        config = config._replace(user_token=ui.censor_token(config.user_token))
    config_to_display = {k: v for k, v in config._asdict().items() if v is not None}
    ui.print_yaml(config_to_display)


app_configure = Typer()


@app_configure.command()
@dagster_cloud_options(allow_empty=True)
def configure(organization: str, deployment: str, api_token: str):
    new_org = ui.input("Dagster Cloud organization:", default=organization or "") or None

    new_api_token = ui.password_input("Dagster Cloud user token:", default=api_token or "") or None

    # Attempt to fetch deployment names from server, fallback to a text input upon failure
    deployment_names = []
    if new_org and new_api_token:
        try:
            deployments = gql.fetch_deployments(
                gql.graphql_client_from_url(gql.url_from_config(new_org), new_api_token)
            )
            deployment_names = [deployment["deploymentName"] for deployment in deployments]
        except:
            ui.warn(
                "Could not fetch deployment names from server - organization or user token may be set incorrectly."
            )

    new_deployment: Optional[str] = None
    if deployment_names:
        options = ["None"] + deployment_names
        new_deployment = ui.list_input(
            "Default deployment:", choices=options, default=deployment or "None"
        )
        if new_deployment == "None":
            new_deployment = None
    else:
        new_deployment = ui.input("Default deployment:", default=deployment or "") or None

    write_config(
        DagsterCloudCliConfig(
            organization=new_org, default_deployment=new_deployment, user_token=new_api_token
        )
    )



import json
from typing import Any, Callable, Dict, List, Optional, Union

import questionary
import typer
import yaml


def as_code(text: str) -> str:
    """
    Returns the given text colored and in backticks.
    """
    return typer.style(f"`{text}`", fg=typer.colors.CYAN)


def red(text: str) -> str:
    return typer.style(text, fg=typer.colors.RED)


def blue(text: str) -> str:
    return typer.style(text, fg=typer.colors.BLUE)


def yellow(text: str) -> str:
    return typer.style(text, fg=typer.colors.YELLOW)


def green(text: str) -> str:
    return typer.style(text, fg=typer.colors.GREEN)


def warn(message: str) -> None:
    """
    Prints a warning message.
    """
    print(f"{yellow('Warning:')} {message}")


def error(message: str, code: int = 1) -> Exception:
    """
    Prints an error message and returns an exception.
    """
    print(f"{red('Error:')} {message}")
    return typer.Exit(code=code)


def censor_token(token: str) -> str:
    return ("*" * (len(token) - 6)) + token[-6:]


def list_input(
    prompt: str,
    choices: List[Any],
    default: Optional[Union[str, questionary.Choice, Dict[str, Any]]] = None,
) -> str:
    """
    Presents the user with a list of choices that can be navigated with
    the up and down arrows.
    """

    return questionary.select(prompt, choices=choices, default=default).ask()


# pylint: disable=redefined-builtin
def input(prompt: str, default: str = "", validate: Callable = None) -> str:
    """
    Prompts the user for text input.
    """

    return questionary.text(prompt, default=default, validate=validate).ask()


def password_input(prompt: str, default: str = "") -> str:
    """
    Prompts the user for text input, hidden by dots.
    """

    return questionary.password(prompt, default=default).ask()


# pylint: disable=redefined-builtin
def print(*args, **kwargs):
    """
    Echos text to the console.
    """
    return typer.echo(*args, **kwargs)


def print_json(data: Any):
    return typer.echo(json.dumps(data, indent=2))


def print_yaml(data: Any):
    return typer.echo(yaml.dump(data))


def erase_previous_line():
    """
    Erases the previous line of output, which can then be printed over.
    """
    return typer.echo("\033[F\033[K\033[F")


def confirm(prompt: str, default: bool = True) -> bool:
    """
    Asks the user to respond yes or no to the prompt.
    """

    return questionary.confirm(prompt, default=default).ask()


def bool_input(prompt: str, default: bool = False) -> bool:
    return confirm(prompt, default=default)



from typer import Typer

from .login import commands as login_cli
from .saml import commands as saml_cli

app = Typer(help="Customize your dagster-cloud settings.")
app.add_typer(login_cli.app, name="login")
app.add_typer(saml_cli.app, name="saml")



from abc import abstractmethod

from dagster.core.launcher import RunLauncher


class WatchfulRunLauncher(RunLauncher):
    @abstractmethod
    def check_run_health(self, run_id):
        """
        Checks the health of a run.

        Args:
            run_id (str): The id of the run to check.
        """
